/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package com.test;

import com.test.ProcessExecutionHandler;
import com.test.ProcessItem;
import java.io.IOException;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import org.apache.log4j.Logger;
import org.drools.KnowledgeBase;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.Resource;
import org.drools.logger.KnowledgeRuntimeLogger;
import org.drools.logger.KnowledgeRuntimeLoggerFactory;
import org.drools.runtime.StatefulKnowledgeSession;

/**
 *
 * @author donaldwalters
 */
public class ExecutionController{

    private static Logger log = Logger.getLogger("executor");
    private Repository repository = null;
    private static final CacheManager manager = CacheManager.create();
    private static final Cache ksessions = new Cache("ksession", 5000, true, false, 60 * 30, 60 * 30, true, 60 * 30);
    private static final AtomicInteger sessionKeys = new AtomicInteger();
    private static final SessionExpirator expirator;

    static {
        manager.addCache(ksessions);
        expirator = new SessionExpirator();
        expirator.start();
    }

    public synchronized static StatefulKnowledgeSession getSession(Integer id) {
        Element e = ksessions.get(id);
        if (e != null) {
            return (StatefulKnowledgeSession) e.getObjectValue();
        }

        return null;
    }

    public synchronized static void removeSession(Integer id) {
        ksessions.remove(id);
    }

    public synchronized static void addSession(Integer id, StatefulKnowledgeSession session) {
        Element e = new Element(id, session);
        ksessions.put(e);
    }

    public void setRepository(Repository repository) {
        this.repository = repository;
    }
    

    public void start(String processName, Map<String, Object> parameters) throws IOException {
        log.debug("starting flow with paams=" + parameters);
        Object[] obj = createKnowledgeBase(processName);
        Integer sessionId = (Integer) obj[0];
        StatefulKnowledgeSession ksession = (StatefulKnowledgeSession) obj[1];

        ProcessItem p = (ProcessItem) parameters.get(Constants.PROCESS_ITEM);
        p.addVariable(Constants.SESSION_ID, sessionId);
        ksession.startProcess("ServiceProcess", parameters);
        ksession.dispose();
    }

    public void resume(String processName, Integer sessionId, String eventSignalRef, ProcessItem eventObject) throws IOException {
        log.debug("going to resume session[id=" + sessionId + ", signalId=" + eventSignalRef + "]");
        log.debug(eventObject);
        eventObject.addVariable(Constants.SESSION_ID, sessionId);
        Object[] obj = loadKnowledgeBase(processName, sessionId, helper);
        StatefulKnowledgeSession ksession = (StatefulKnowledgeSession) obj[1];
        ksession.signalEvent(eventSignalRef, eventObject);
    }

    private Object[] createKnowledgeBase(String processName) throws IOException {
        String filename = processName + ".bpmn2";
        log.debug("retrieving execution process " + processName);

        Resource resource = repository.getResource(filename);
        log.debug("creating knowledge builder");
        KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();

        kbuilder.add(resource, ResourceType.BPMN2);
        log.debug("creating knowledge base");

        KnowledgeBase kbase = kbuilder.newKnowledgeBase();

        StatefulKnowledgeSession ksession = kbase.newStatefulKnowledgeSession();//CacheKnowledgeService.newStatefulKnowledgeSession(kbase);

        if (log.isTraceEnabled()) {
            KnowledgeRuntimeLogger logger =
                    KnowledgeRuntimeLoggerFactory.newConsoleLogger(ksession);
        }

        ksession.getWorkItemManager().registerWorkItemHandler(
                "Service Task", new ProcessExecutionHandler());

//        ksessions.put(ksession.getId(), ksession);

        Integer key = sessionKeys.incrementAndGet();
        Object[] obj = new Object[2];
        obj[0] = key;
        obj[1] = ksession;


        addSession(key, ksession);

        return obj;
    }

    private Object[] loadKnowledgeBase(String processName, int sessionId) throws IOException {

        if (sessionId == 0) {
            return createKnowledgeBase(processName, helper);
        }

        String filename = processName + ".bpmn2";
        log.debug("retrieving execution process " + processName);

        Resource resource = repository.getResource(filename);
        log.debug("creating knowledge builder");
        KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();

        kbuilder.add(resource, ResourceType.BPMN2);
        log.debug("creating knowledge base");

        KnowledgeBase kbase = kbuilder.newKnowledgeBase();
        log.debug("creating session " + sessionId);


        StatefulKnowledgeSession ksession = getSession(sessionId);
        //ksessions.get(sessionId);//CacheKnowledgeService.loadStatefulKnowledgeSession(sessionId, kbase);

        log.debug("getSession " + ksession);

        if (log.isTraceEnabled()) {
            KnowledgeRuntimeLogger logger =
                    KnowledgeRuntimeLoggerFactory.newConsoleLogger(ksession);
        }

        ksession.getWorkItemManager().registerWorkItemHandler(
                "Service Task", new ProcessExecutionHandler());

        Object[] obj = new Object[2];
        obj[0] = sessionId;
        obj[1] = ksession;

        return obj;

    }

    private static class SessionExpirator extends Thread {

        @Override
        public void run() {
            while (RunningStatus.isRunning()) {
                for (Iterator keys = ksessions.getKeys().iterator(); keys.hasNext();) {
                    StatefulKnowledgeSession sess = getSession((Integer) keys.next());
                    sess.fireUntilHalt();
                }
            }
            log.debug(ksessions.getSize() + " sessions in memory");

            try {
                Thread.sleep(1000 * 60);
            } catch (Exception e) {
            }
        }
    }
}
